# This script has been tested on Windows 10 with Python 3.11

import os
import tempfile
import pyarrow.parquet as pq
from azure.storage.blob import BlobServiceClient
import pymssql

# Initialize the BlobServiceClient
def initialize_blob_client(connection_string_output, container_output):
    blob_service_client = BlobServiceClient.from_connection_string(connection_string_output)
    return blob_service_client.get_container_client(container_output)

# List all device/message folders in the container
def list_device_message_folders(container_client):
    blobs = container_client.list_blobs()
    device_message_folders = set()
    
    for blob in blobs:
        parts = blob.name.split('/')
        if len(parts) == 5:  # device/message/yyyy/mm/dd
            device_message_folders.add('/'.join(parts[:2]))
    
    return list(device_message_folders)

# Sample the first Parquet file to get the schema
def get_parquet_schema(container_client, folder_path):
    blobs = container_client.list_blobs(name_starts_with=folder_path)
    
    for blob in blobs:
        if blob.name.endswith('.parquet'):
            blob_client = container_client.get_blob_client(blob)
            downloaded_blob = blob_client.download_blob().readall()
            
            with tempfile.NamedTemporaryFile(delete=False) as temp_file:
                temp_file.write(downloaded_blob)
                temp_file_path = temp_file.name

            try:
                table = pq.read_table(temp_file_path)
            finally:
                os.remove(temp_file_path)
                
            return table.schema
    return None

# Generate the CREATE EXTERNAL TABLE SQL statement
def generate_create_external_table_sql(table_name, schema, location):
    columns = []
    for field in schema:
        if field.name == "t":
            column = f"[{field.name}] DATETIME"
        else:
            column = f"[{field.name}] FLOAT"
        columns.append(column)
        
    columns_sql = ",\n\t".join(columns)
    sql = f"""
    CREATE EXTERNAL TABLE [tbl_{table_name}]
    (
        {columns_sql}
    )
    WITH
    (
        LOCATION = '{location}/*',
        DATA_SOURCE = [ParquetDataLake],
        FILE_FORMAT = [ParquetFormat]
    )
    """
    return sql

# Drop the external table if it exists
def drop_external_table_if_exists(cursor, table_name):
    drop_sql = f"""
    IF EXISTS (SELECT * FROM sys.external_tables WHERE name = 'tbl_{table_name}')
    DROP EXTERNAL TABLE [tbl_{table_name}]
    """
    cursor.execute(drop_sql)

# Create or overwrite the external table in Azure Synapse
def create_external_table(cursor, sql, table_name):
    try:
        drop_external_table_if_exists(cursor, table_name)
        print(f"Executing SQL for table: tbl_{table_name}")
        cursor.execute(sql)
    except pymssql.Error as e:
        print(f"Error creating table tbl_{table_name}: {e}")

def create_database_and_objects(synapse_server, synapse_user, synapse_password, synapse_database, storage_account_name, container_output, master_key_password):
    # Connect to the server (not to any specific database)
    conn = pymssql.connect(server=synapse_server, user=synapse_user, password=synapse_password, database="master")
    conn.autocommit(True)
    cursor = conn.cursor()

    # Check if the database exists
    cursor.execute(f"SELECT COUNT(*) FROM sys.databases WHERE name = '{synapse_database}'")
    if cursor.fetchone()[0] == 0:
        # Create database
        cursor.execute(f"CREATE DATABASE {synapse_database}")
        print(f"Database {synapse_database} created.")

    cursor.close()
    conn.close()

    # Connect to the new database to create necessary objects
    conn = pymssql.connect(server=synapse_server, user=synapse_user, password=synapse_password, database=synapse_database)
    cursor = conn.cursor()

    # Check and create necessary objects
    cursor.execute(f"SELECT COUNT(*) FROM sys.symmetric_keys WHERE name = '##MS_DatabaseMasterKey##'")
    if cursor.fetchone()[0] == 0:
        cursor.execute(f"CREATE MASTER KEY ENCRYPTION BY PASSWORD = '{master_key_password}'")

    cursor.execute(f"SELECT COUNT(*) FROM sys.database_scoped_credentials WHERE name = 'my_credential'")
    if cursor.fetchone()[0] == 0:
        cursor.execute(f"CREATE DATABASE SCOPED CREDENTIAL my_credential WITH IDENTITY = 'Managed Identity'")

    cursor.execute(f"SELECT COUNT(*) FROM sys.external_data_sources WHERE name = 'ParquetDataLake'")
    if cursor.fetchone()[0] == 0:
        cursor.execute(f"CREATE EXTERNAL DATA SOURCE ParquetDataLake WITH (LOCATION = 'https://{storage_account_name}.dfs.core.windows.net/{container_output}', CREDENTIAL = my_credential)")

    cursor.execute(f"SELECT COUNT(*) FROM sys.external_file_formats WHERE name = 'ParquetFormat'")
    if cursor.fetchone()[0] == 0:
        cursor.execute(f"CREATE EXTERNAL FILE FORMAT ParquetFormat WITH (FORMAT_TYPE = PARQUET)")

    conn.commit()
    print(f"Database {synapse_database} and objects created.")

    cursor.close()
    conn.close()

def main(connection_string_output, container_output, synapse_server, synapse_database, synapse_user, synapse_password, master_key_password, storage_account_name):
    container_client = initialize_blob_client(connection_string_output, container_output)
    folders = list_device_message_folders(container_client)
    
    # Create database and necessary objects if they do not exist
    create_database_and_objects(synapse_server, synapse_user, synapse_password, synapse_database, storage_account_name, container_output, master_key_password)
    
    # Connect to Azure Synapse using pymssql and switch to the created database
    conn = pymssql.connect(server=synapse_server, user=synapse_user, password=synapse_password, database=synapse_database)
    cursor = conn.cursor()
    
    for folder in folders:
        schema = get_parquet_schema(container_client, folder)
        if schema is not None:
            table_name = folder.replace('/', '_')
            location = f'/{folder}'
            sql = generate_create_external_table_sql(table_name, schema, location)
            create_external_table(cursor, sql, table_name)
            conn.commit()
    
    cursor.close()
    conn.close()

    
if __name__ == "__main__":
    # update below variables
    storage_account_output = "your-output-storage-account" # e.g. parquetstorageaccount
    container_output = "your-output-container-name" # e.g. parquet-container-output
    connection_string_output = "your-output-connection-string" # find this in your output storage account under 'Security + networking/Access keys/Connection string'
    synapse_server = "your-serverless-sql-endpoint" # find this from your Synapse workspace 'Overview' page
    synapse_password = "your-synapse-sql-admin-user-password" # this was specified in the 'Network' section when creating and can optionally be reset in the Synapse workspace overview 
    master_key_password = "yourVerystrongpassword12345@*" # specify a master key - make sure that it is sufficiently strong, else the script fails
    
    # keep these variables as-is
    synapse_database = "parquetdatalake"
    synapse_user = "sqladminuser"
    
    main(connection_string_output, container_output, synapse_server, synapse_database, synapse_user, synapse_password, master_key_password, storage_account_output)
